#include "LCD_Optrex.h"

#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include "Arduino.h"

// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set: 
//    DL = 1; 8-bit interface data 
//    N = 0; 1-line display 
//    F = 0; 5x8 dot character font 
// 3. Display on/off control: 
//    D = 0; Display off 
//    C = 0; Cursor off 
//    B = 0; Blinking off 
// 4. Entry mode set: 
//    I/D = 1; Increment by 1 
//    S = 0; No shift 
//
// Note, however, that resetting the Arduino doesn't reset the LCD, so we
// can't assume that its in that state when a sketch starts (and the
// LCD_Optrex constructor is called).

LCD_Optrex::LCD_Optrex(uint8_t rs, uint8_t rw, uint8_t enable,
  uint8_t d0, uint8_t d1, uint8_t d2, uint8_t d3,
  uint8_t d4, uint8_t d5, uint8_t d6, uint8_t d7) :
  _four_bit_mode(0), _rs_pin(rs), _rw_pin(rw), _enable_pin(enable)
{
  _data_pins[0] = d0;
  _data_pins[1] = d1;
  _data_pins[2] = d2;
  _data_pins[3] = d3; 
  _data_pins[4] = d4;
  _data_pins[5] = d5;
  _data_pins[6] = d6;
  _data_pins[7] = d7; 
  
  pinMode(_rs_pin, OUTPUT);
  pinMode(_rw_pin, OUTPUT);
  pinMode(_enable_pin, OUTPUT);
  
  for (int i = 0; i < 8; i++)
    pinMode(_data_pins[i], OUTPUT);

	delayMicroseconds(16000);	// mandatory delay for Vcc stabilization
	command(0x30);				// set 8-bit mode
	delayMicroseconds(5000);	// mandatory delay
	command(0x30);
	delayMicroseconds(200);
	command(0x30);
	command(0x38);			// 8-bit, 2-line, 5x7 char set
	command(0x08);			// display off
	command(0x01);			// clear display
	delayMicroseconds(16000);
	command(0x06);			// increment, no shift
	command(0x0c);			// display on, no cursor, no blink

}

LCD_Optrex::LCD_Optrex(uint8_t rs, uint8_t rw, uint8_t enable,
  uint8_t d0, uint8_t d1, uint8_t d2, uint8_t d3) :
  _four_bit_mode(1), _rs_pin(rs), _rw_pin(rw), _enable_pin(enable)
{
  _data_pins[0] = d0;
  _data_pins[1] = d1;
  _data_pins[2] = d2;
  _data_pins[3] = d3; 
  
  pinMode(_rs_pin, OUTPUT);
  pinMode(_rw_pin, OUTPUT);
  pinMode(_enable_pin, OUTPUT);
  
  for (int i = 0; i < 4; i++)
    pinMode(_data_pins[i], OUTPUT);

	delayMicroseconds(16000);		// mandatory delay for Vcc stabilization
	sendraw4(0x30);					// set 8-bit mode (yes, it's true!)
	delayMicroseconds(5000);		// mandatory delay
	sendraw4(0x30);
	delayMicroseconds(200);
	sendraw4(0x30);
	delayMicroseconds(40);			// command delay
	sendraw4(0x20);					// finally set 4-bit mode
	delayMicroseconds(40);			// command delay

	command(0x28);			// 4-bit, 2-line, 5x7 char set
	command(0x08);			// display off
	command(0x01);			// clear display
	delayMicroseconds(16000);
	command(0x06);			// increment, no shift
	command(0x0c);			// display on, no cursor, no blink
}

void LCD_Optrex::clear()
{
  command(0x01);  // clear display, set cursor position to zero
  delayMicroseconds(16000);
}

void LCD_Optrex::home()
{
  command(0x02);  // set cursor position to zero
  delayMicroseconds(16000);
}

void LCD_Optrex::setCursor(int col, int row)
{
  int row_offsets[] = { 0x00, 0x40, 0x14, 0x54 };
  command(0x80 | (col + row_offsets[row]));
}

// These commands scroll the display without changing the RAM
void LCD_Optrex::scrollDisplayLeft(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}
void LCD_Optrex::scrollDisplayRight(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

// This is for text that flows Left to Right
void LCD_Optrex::leftToRight(void) {
  _displaymode |= LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This is for text that flows Right to Left
void LCD_Optrex::rightToLeft(void) {
  _displaymode &= ~LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// Allows us to fill the first 8 CGRAM locations
void LCD_Optrex::createChar(uint8_t location, uint8_t charmap[]) {
// with custom characters
  location &= 0x7; // we only have 8 locations 0-7
  command(LCD_SETCGRAMADDR | (location << 3));
  for (int i=0; i<8; i++) {
    write(charmap[i]);
  }
}

void LCD_Optrex::command(uint8_t value) {
  send(value, LOW);
}

size_t LCD_Optrex::write(uint8_t value) {
  send(value, HIGH);
  return 1;
}

void LCD_Optrex::send(uint8_t value, uint8_t mode) {
  digitalWrite(_rs_pin, mode);
  digitalWrite(_rw_pin, LOW);

  if (_four_bit_mode) {
    for (int i = 0; i < 4; i++) {
      digitalWrite(_data_pins[i], (value >> (i + 4)) & 0x01);
    }
    
    digitalWrite(_enable_pin, HIGH);
    digitalWrite(_enable_pin, LOW);
    
    for (int i = 0; i < 4; i++) {
      digitalWrite(_data_pins[i], (value >> i) & 0x01);
    }

    digitalWrite(_enable_pin, HIGH);
    digitalWrite(_enable_pin, LOW);
  } else {
    for (int i = 0; i < 8; i++) {
      digitalWrite(_data_pins[i], (value >> i) & 0x01);
    }

    digitalWrite(_enable_pin, HIGH);
    digitalWrite(_enable_pin, LOW);
  }

  delayMicroseconds(40);
}

void LCD_Optrex::sendraw4(uint8_t value) {
  digitalWrite(_rs_pin, LOW);
  digitalWrite(_rw_pin, LOW);

    for (int i = 0; i < 4; i++) {
      digitalWrite(_data_pins[i], (value >> (i + 4)) & 0x01);
    }

    digitalWrite(_enable_pin, HIGH);
    digitalWrite(_enable_pin, LOW);
}
